'''Python bindings for the Sophos Anti-Virus Interface (SAVI)

Types:

  Savi              An alias for the most recent 'Savi' type available in
                    pysavi.  Currently another name for 'Savi3'.

  Savi3             An instance of the SAVI engine.  It is the only class with
                    a user-accessible constructor, and may be subclassed.

  EnumGeneric       Basic iterator for many of the types listed below,
                    returned by various ISavi3 methods.

  SweepResults      Represents a single, specific virus found during a scan.

  SweepError        Contains details of errors encountered during scanning or
                    disinfection.

  EngineConfig      Represents a single configuration option's name and type.

  IDEDetailsType    Represents an update file in use by the engine.

  VersionChecksum   Represents a checksum of an engine component.

Exceptions:

  error             Carries an integral error code and descriptive string.
                    Subclass of EnvironmentError.

Functions and Constants:

  Too numerous to list here.  Aside from error codes and SAVI-specific
  constants, there are functions to test success values of error codes
  (SOPHOS_SUCCEEDED, SOPHOS_FAILED) and to return human-friendly strings
  describing error codes (strerror).

User-defined Classes:

  SAVI streams          Any object offering file-alike seek(), tell(), read()
                        and, if disinfection is desired, write() methods may
                        be passed to the ISavi3 ...Handle() methods.  StringIO
                        objects offer this interface, for instance.  The user
                        may implement arbitrarily sophisticated data-access
                        schemes with this approach that do not require
                        conventional files.

  SAVI callbacks        Any object offering methods named after the SAVI
                        callback interface methods (OnVirusFound, OnChange,
                        etc.) may serve as a callback dispatcher for an ISavi3
                        instance via RegisterNotification().

'''

# $Id: __init__.py,v 1.4 2005/02/08 06:22:16 mjp Exp $

# classes/types
from _savi import *

# misc SAVI API functions
def SOPHOS_SUCCEEDED(Status):  return Status >= 0
def SOPHOS_FAILED(Status):     return Status < 0

# vim:et:ts=2:noai:
