/**
pysavi 0.30

Copyright 2002, 2003, 2004 Michael J. Pomraning
  <mjp*DASH*pysavi*AT*pilcrow*DOT*madison*DOT*wi*DOT*us>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/

/* $Id: _savimodule.c,v 1.4 2005/03/11 03:20:49 mjp Exp $ */

/* _savimodule.c */

#include <Python.h>
#include "csavi3c.h"
#include "pysavi.h"
#include "pysaviUtil.h"
#include "pysaviTypes.h"

#define VERSION     "0.31"
#define REVISION    "$Id: _savimodule.c,v 1.4 2005/03/11 03:20:49 mjp Exp $"

/* === Objects and Types =================================================== */

PyObject * pysErrorObject;

staticforward PyTypeObject PySSavi3_Type;

/* === Module Functions ==================================================== */

static char
pysavi_strerror__doc__[] =
"strerror(code) -> string\n"
"\n"
"Return a string describing the error code.\n";

static PyObject *
pysavi_strerror(PyObject *dummy, PyObject *args) {
  HRESULT hr;

  if (!PyArg_ParseTuple(args, "i:strerror", (long *)&hr)) return NULL;
  return PyString_FromString(pysStrerror(hr));
}

/* === PySSavi3 Object Implementation ====================================== */

static char PySSavi3__doc__[] =
"Savi3(moniker=None)\n"
"\n"
"Create a new Savi3 object for virus scanning and disinfection.  If given,\n"
"the moniker argument is passed to and applied during the object's __init__\n"
"method.  Monikers are only needed on Windows platforms.\n"
"\n"
"Creation does not load any virus data; see the LoadVirusData method\n";

static PyObject *
pys_I_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds) {
  HRESULT hr;
  CISweepClassFactory2 *pFactory = NULL;
  PySSavi3 *self = NULL;
  PyThreadState *tstate = NULL;
  
  self = (PySSavi3 *)subtype->tp_alloc(subtype, 0);
  if (self == NULL) return (PyObject *)self;

  self->pSAVI = NULL;
  self->cbck  = NULL;
  tstate = PyThreadState_Get(); /* THREAD! */
  self->interp = tstate->interp;

  hr = DllGetClassObject((REFCLSID)&SOPHOS_CLASSID_SAVI,
                         (REFIID)&SOPHOS_IID_CLASSFACTORY2,
                         (void **) &pFactory);
  if (SOPHOS_FAILED(hr)) goto savierr;
  hr = pFactory->pVtbl->CreateInstance(pFactory,
                                       NULL, &SOPHOS_IID_SAVI3,
                                       (void **)&((PySSavi3 *)self)->pSAVI);
  if (SOPHOS_FAILED(hr)) goto savierr;

out:
  pysIThingSafeRelease(pFactory);
  return (PyObject *)self;
  
savierr:
  if (self) {
    pysIThingSafeRelease(((PySSavi3 *)self)->pSAVI);
    PyMem_DEL(self);
  }
  pysRaiseError(hr);
  self = NULL;
  goto out;
}

static int
pys_I_init(PySSavi3 *self, PyObject *args, PyObject *kwds) {
  HRESULT hr;
  static char *kwlist[] = {"moniker", NULL};
  char * moniker = NULL;

  if (!PyArg_ParseTupleAndKeywords(args, kwds,
                   "|z:Savi3", kwlist, &moniker))
    return -1;

  if (NULL == moniker) {
    Py_BEGIN_ALLOW_THREADS
    hr = self->pSAVI->pVtbl->Initialise(self->pSAVI);
    Py_END_ALLOW_THREADS
  } else {
    Py_BEGIN_ALLOW_THREADS
    hr = self->pSAVI->pVtbl->InitialiseWithMoniker(self->pSAVI, moniker);
    Py_END_ALLOW_THREADS
  }

  if (!SOPHOS_SUCCEEDED(hr)) {
    return -1;
    pysRaiseError(hr);
  }
  return 0;
}

static void
pys_I_dealloc(PySSavi3 *self) {
  /* callback de-registration is implied by pSAVI dismissal */

  if (self->pSAVI) {
    self->pSAVI->pVtbl->Terminate(self->pSAVI);
    pysIThingSafeRelease(self->pSAVI);
  }

  self->ob_type->tp_free((PyObject *)self);
}

static char pys_I_RegisterNotification__doc__[] =
"RegisterNotification(cbck)\n"
"\n"
"Register the cbck object as a callback dispatcher for this SAVI engine\n"
"instance.  The cbck object must implement at least one SAVI callback method:\n"
"OnVirusFound, OnFileFound, OnErrorFound, OkToContinue, OnClassification,\n"
"OnChange or OnSevereError.\n"
"\n"
"It is acceptable to register oneself as a callback dispatcher, likely in the\n"
"__init__ routine of a subclass.\n";

static PyObject *
pys_I_RegisterNotification(PySSavi3 *self, PyObject *args) {
/*
 * RegisterNotification(broker = None):
 *
 * If broker is None, de-register all callbacks.
 * 
 * Otherwise, set up CISweepNotify with the *self object as the token passed
 * to callback functions.  *self holds *broker, which, in turn, implements all
 * callbacks as methods ((OnFileFound, OnVirusFound, OnErrorFound).  pNotify
 * pointer is also stored under *self.
 *
 *  args:  None or broker object (None clears current registration)
 *
 */
  int i;
  static REFIID cbckIIDs[] = { &SOPHOS_IID_SWEEPNOTIFY2,
                                &SOPHOS_IID_CHANGENOTIFY,
                                &SOPHOS_IID_SEVERITYNOTIFY };
  static char *SweepNotifyTbl[] =
  {"OnVirusFound","OnFileFound","OnErrorFound","OkToContinue",
   "OnClassification"};
  PyObject *broker = Py_None;

  if (!PyArg_ParseTuple(args, "|O:sweepnotify", &broker))
    return NULL;

  if (self->cbck) {
    for (i = 0; i < sizeof(cbckIIDs)/sizeof(cbckIIDs[0]); i++) {
      (void) self->pSAVI->pVtbl->RegisterNotification(self->pSAVI,
                                                      cbckIIDs[i], NULL, NULL);
    }
    Py_DECREF(self->cbck);
    self->cbck = NULL;
  }

  if (broker == Py_None) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  for (i = 0; i < sizeof(SweepNotifyTbl)/sizeof(SweepNotifyTbl[0]); i++) {
    if (pysPyObjHasMethod(broker, SweepNotifyTbl[i])) {
      self->cbck = broker;
      if (!pysCbckRegister(self, &SOPHOS_IID_SWEEPNOTIFY2))
        return NULL;
      break;
    }
  }

  /* FIXME - deregister callbacks on failure */

  if (pysPyObjHasMethod(broker, "OnChange")) {
    self->cbck = broker;
    if (!pysCbckRegister(self, &SOPHOS_IID_CHANGENOTIFY))
      return NULL;
  }

  if (pysPyObjHasMethod(broker, "OnSevereError")) {
    self->cbck = broker;
    if (!pysCbckRegister(self, &SOPHOS_IID_SEVERITYNOTIFY))
      return NULL;
  }

  if (!self->cbck) {
    PyErr_SetString(PyExc_TypeError,
                    "callback object implemented no known SAVI callbacks");
    return NULL;
  }

  if ((void *)broker != (void *)self) {
    Py_INCREF(broker);
  }

  Py_INCREF(Py_None);
  return Py_None;

}

static char pys_I_SetConfigDefaults__doc__[] =
"SetConfigDefaults()\n"
"\n"
"Reset all configuration options to their defaults.\n";

static PyObject *
pys_I_SetConfigDefaults(PySSavi3 *self) { /* METH_NOARGS */
  HRESULT hr;

  hr = self->pSAVI->pVtbl->SetConfigDefaults(self->pSAVI);
  if (SOPHOS_SUCCEEDED(hr)) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  return pysRaiseError(hr);
}

static char pys_I_GetConfigEnumerator__doc__[] =
"GetConfigEnumerator() -> engine_cfg\n"
"\n"
"Return an iterator of EngineConfigType objects, which may be used to list\n"
"the names and integral type codes of all configuration options.  To query\n"
"or set the value of an individual option, see GetConfigValue and\n"
"SetConfigValue.\n";

static PyObject *
pys_I_GetConfigEnumerator(PySSavi3 *self) {
  HRESULT hr;
  PyObject *eec_obj;
  CIEnumEngineConfig *pConfigEnum = NULL;

  hr = self->pSAVI->pVtbl->GetConfigEnumerator(
         self->pSAVI,
         &SOPHOS_IID_ENUM_ENGINECONFIG,
         (void **)&pConfigEnum);
  if (!SOPHOS_SUCCEEDED(hr)) {
    pysIThingSafeRelease(pConfigEnum);
    return pysRaiseError(hr);
  }

  eec_obj = PySEnumGeneric_New(&SOPHOS_IID_ENUM_ENGINECONFIG,
                               (CIEnumGeneric *)pConfigEnum, 1);
  if (eec_obj == NULL)
    pysIThingSafeRelease(pConfigEnum);

  return eec_obj;
}

static char pys_I_SetConfigValue__doc__[] =
"SetConfigValue(name, type, data)\n"
"\n"
"Set the named option, of the given type, to the value specified in data.\n"
"Note that data values are given as strings.  For example:\n"
"\n"
"  savi.SetConfigValue('GrpArchiveUnpack', SOPHOS_TYPE_OPTION_GROUP, '1')\n";

static PyObject *
pys_I_SetConfigValue(PySSavi3 *self, PyObject *args) {
  char * config, * value;
  int s_type;
  HRESULT hr;

  if (!PyArg_ParseTuple(args, "sis:SetConfigValue", &config, &s_type, &value))
    return NULL;

  hr = self->pSAVI->pVtbl->SetConfigValue(
       self->pSAVI, config, (U32) s_type, value);

  if (!hr == SOPHOS_S_OK)
    return pysRaiseError(hr);

  Py_INCREF(Py_None);
  return Py_None;
}

static char pys_I_GetConfigValue__doc__[] =
"GetConfigValue(name, type) -> value\n"
"\n"
"Get the value of the named option, of the given type, as a string.  This\n"
"method cannot list all available options (but see GetConfigEnumerator).\n";

static PyObject *
pys_I_GetConfigValue(PySSavi3 *self, PyObject *args) {
  char *config;
  int s_type;
  OLECHAR value[256];
  HRESULT hr;

  if (!PyArg_ParseTuple(args, "si:GetConfigValue", &config, &s_type))
    return NULL;

  /* pysIThingGetString doesn't handle type */
  hr = self->pSAVI->pVtbl->GetConfigValue(self->pSAVI,
           config, (U32) s_type, sizeof(value)-1, value, NULL);

  if (!SOPHOS_SUCCEEDED(hr))
    return pysRaiseError(hr);

  return PyString_FromString(value);
}

/*
 * _pysavi_postSweep(PySSavi3, HRESULT, pEnumResults)
 *
 * Translate the outcome of Sweep...() calls (HRESULT and possible
 * CIEnumSweepResults), checking for occurrence of errors in callbacks, which
 * errors are stored in the PySSavi3 object.
 *
 * SOPHOS_S_OK.................. boolean False
 *                               -or-
 *                               SweepResultsType (zero members)
 *
 * VIRUSPRESENT................. boolean True
 *                               -or-
 *                               SweepResultsType (>=1 members)
 *
 * exception in callback........ n/a (re-raise exception)
 *                               release pEnumResults, clear exc_*
 *                               "exception storage" members
 */
static PyObject *
_pysavi_postSweep(PySSavi3 *self, HRESULT hr, CIEnumSweepResults *pEnumResults) {
  PyObject *ret = NULL;
  int virus_present = 0;

  if (PyErr_Occurred()) goto failed; /* ??? Never true ??? */
  if (self->exc_type) {
    PyErr_Restore(self->exc_type, self->exc_val, self->exc_tb);
    goto failed;
  }

  if (!SOPHOS_SUCCEEDED(hr)) {
    pysRaiseError(hr);
    goto failed;
  }

  virus_present = (hr == SOPHOS_SAVI_ERROR_VIRUSPRESENT);
  
  if (!pEnumResults) {                         /* boolean return */
    ret = virus_present ? Py_True : Py_None;
    Py_INCREF(ret);
  } else {                                     /* EnumSweepResults return */
    ret = PySEnumGeneric_New(&SOPHOS_IID_ENUM_SWEEPRESULTS,
                             pEnumResults, virus_present);
    if (ret == NULL) goto failed;
  }

  return ret;

failed:
  pysIThingSafeRelease(pEnumResults);
  self->exc_type = self->exc_val = self->exc_tb = NULL;
  return NULL;
}

/* 
 * results_converter()
 *
 * PyArg_Parse converter function, mapping the user-supplied "results" param
 * of the Sweep...()/Disinfect...() functions to the desired return value:
 *
 * input parameter     output parameter     meaning
 * ---------------     ----------------     -------
 * PySSweepResults           1              SweepResults iterator desired
 * Py_None                   0              boolean desired
 * 
 */

static int
results_converter(PyObject *o, void *want_iter) {
  if (o == (PyObject *)&PySSweepResults_Type) {
    *(int *)want_iter = 1;
  } else if (o == Py_None) {
    *(int *)want_iter = 0;
  } else {
    PyErr_SetString(PyExc_TypeError,
                    "results parameter must be SweepResults or None");
    return 0;
  }
  return 1;
}

#define PYS_FUNCFILE(pfx)                                                 \
static PyObject *                                                         \
pys_I_##pfx##File(PySSavi3 *self, PyObject *args, PyObject *kwds) {       \
  static char *kwlist[] = { "file", "results", NULL };                    \
  char *file = NULL;                                                      \
  int want_sr = 1;                                                        \
  CIEnumSweepResults *pResults = NULL;                                    \
  HRESULT hr;                                                             \
                                                                          \
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|O&:" #pfx "File",       \
                                   kwlist, &file, results_converter,      \
                                   &want_sr))                             \
    return NULL;                                                          \
                                                                          \
  Py_BEGIN_ALLOW_THREADS                                                  \
  hr = self->pSAVI->pVtbl->pfx##File(                                     \
          self->pSAVI,                                                    \
          file,                                                           \
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                          \
          want_sr ? (void **)&pResults : NULL);                           \
  Py_END_ALLOW_THREADS                                                    \
                                                                          \
  return _pysavi_postSweep(self, hr, pResults);                           \
}

PYS_FUNCFILE(Sweep)
static char pys_I_SweepFile__doc__[] = 
"SweepFile(file, results=SweepResults) -> viruses\n"
"\n"
"Sweep the specified file (a pathname) for viruses.  May raise an exception,\n"
"and will invoke callbacks as needed.\n"
"\n"
"Return value is by default an iterator of SweepResults objects, each element\n"
"representing a virus found -- see SweepResults.__doc__.  If None is supplied\n"
"as the results parameter SweepFile returns a simple boolean instead.  In\n"
"either case, the return value evaluates to boolean true if any viruses were\n"
"found, false otherwise.\n";

PYS_FUNCFILE(Disinfect)
static char pys_I_DisinfectFile__doc__[] =
"DisinfectFile(file, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the specified file, and then scan for remaining\n"
"viruses.  The return value, which is controlled as in SweepFile, represents\n"
"viruses that remain in the object, i.e., that avoided disinfection.\n";

#define PYS_FUNCLOGICALSECTOR(pfx)                                            \
static PyObject *                                                             \
pys_I_##pfx##LogicalSector(PySSavi3 *self, PyObject *args, PyObject *kwds) {  \
  static char *kwlist[] = { "drive", "sector", "results", NULL };             \
  char *drive = NULL;                                                         \
  U32 sector;                                                                 \
  int want_sr = 1;                                                            \
  CIEnumSweepResults *pResults = NULL;                                        \
  HRESULT hr;                                                                 \
                                                                              \
  if (!PyArg_ParseTupleAndKeywords(args, kwds,                                \
                                   "si|O&:" #pfx "LogicalSector",             \
                                   kwlist, &drive, &sector,                   \
                                   results_converter, &want_sr))              \
    return NULL;                                                              \
                                                                              \
  Py_BEGIN_ALLOW_THREADS                                                      \
  hr = self->pSAVI->pVtbl->pfx##LogicalSector(                                \
         self->pSAVI, (LPCOLESTR)drive,                                       \
         0, sector,                                                           \
         (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                               \
         want_sr ? (void **)&pResults : NULL);                                \
  Py_END_ALLOW_THREADS                                                        \
                                                                              \
  return _pysavi_postSweep(self, hr, pResults);                               \
}

PYS_FUNCLOGICALSECTOR(Sweep)
static char pys_I_SweepLogicalSector__doc__[] =
"DisinfectLogicalSector(drive, sector, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the drive sector.  The sector parameter is zero-based.\n"
"\n"
"Return value is governed by the results parameter, as for SweepFile.\n";

PYS_FUNCLOGICALSECTOR(Disinfect)
static char pys_I_DisinfectLogicalSector__doc__[] =
"DisinfectLogicalSector(drive, sector, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the drive sector.  The sector parameter is zero-based.\n"
"\n"
"Return value is governed by the results parameter, as for DisinfectFile.\n";
        
#define PYS_FUNCPHYSICALSECTOR(pfx)                                           \
static PyObject *                                                             \
pys_I_##pfx##PhysicalSector(PySSavi3 *self, PyObject *args, PyObject *kwds) { \
  static char *kwlist[] = { "drive", "head", "cylinder",                      \
                            "sector", "results", NULL };                      \
  LPCOLESTR drive;                                                            \
  U32 head, cyl, sector;                                                      \
  int want_sr = 1;                                                            \
  CIEnumSweepResults *pResults = NULL;                                        \
  HRESULT hr;                                                                 \
                                                                              \
  if (!PyArg_ParseTupleAndKeywords(args, kwds,                                \
                                   "siii|O&:" #pfx "PhysicalSector",          \
                                   kwlist,                                    \
                                   &drive, &head, &cyl, &sector,              \
                                   results_converter, &want_sr))              \
    return NULL;                                                              \
                                                                              \
  Py_BEGIN_ALLOW_THREADS                                                      \
  hr = self->pSAVI->pVtbl->pfx##PhysicalSector(                               \
          self->pSAVI,                                                        \
          drive, head, cyl, sector,                                           \
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                              \
          want_sr ? (void **)&pResults : NULL);                               \
  Py_END_ALLOW_THREADS                                                        \
                                                                              \
  return _pysavi_postSweep(self, hr, pResults);                               \
}

PYS_FUNCPHYSICALSECTOR(Sweep)
static char pys_I_SweepPhysicalSector__doc__[] =
"SweepPhysicalSector(drive, head, cylinder, sector, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the drive sector.  The head and cylinder parameters are\n"
"zero-indexed, the sector parameter one-indexed.  Return value is governed by\n"
"the results parameter, as for SweepFile.\n";

PYS_FUNCPHYSICALSECTOR(Disinfect)
static char pys_I_DisinfectPhysicalSector__doc__[] =
"DisinfectPhysicalSector(drive, head, cylinder, sector, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the drive sector.  The head and cylinder parameters are\n"
"zero-indexed, the sector parameter one-indexed.  Return value is governed by\n"
"the results parameter, as for DisinfectFile.\n";

static char pys_I_SweepMemory__doc__[] =
"SweepMemory(results=SweepResults) -> viruses\n"
"\n"
"Sweep executable segments of running processes for viruses.  Return value\n"
"is governed by the results parameter, as for SweepFile.\n";

static PyObject *
pys_I_SweepMemory(PySSavi3 *self, PyObject *args, PyObject *kwds) {
  static char *kwlist[] = { "results", NULL };
  int want_sr = 1;
  CIEnumSweepResults *pResults = NULL;
  HRESULT hr;

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&:SweepMemory",
                                   kwlist, results_converter,
                                   &want_sr))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  hr = self->pSAVI->pVtbl->SweepMemory(
          self->pSAVI,
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,
          want_sr ? (void **)&pResults : NULL);
  Py_END_ALLOW_THREADS

  return _pysavi_postSweep(self, hr, pResults);
}

static char pys_I_Disinfect__doc__[] =
"Disinfect(SweepResults_object)\n"
"\n"
"Attempt to disinfect a specific virus from a specific item, represented by\n"
"the SweepResults object returned by any of the Sweep...() methods.  Raises\n"
"an exception if disinfection was not completely successful.\n"
"\n"
"Even when successful, other viruses may remain in the underlying object.\n";

static PyObject *
pys_I_Disinfect(PySSavi3 *self, PyObject *args) {
  HRESULT hr;
  CISweepResults *sr = NULL;

  if (!PyArg_ParseTuple(args, "O&:Disinfect", PySSweepResults_converter, &sr))
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  hr = self->pSAVI->pVtbl->Disinfect(
          self->pSAVI,
          (REFIID)&SOPHOS_IID_SWEEPRESULTS,
          (void *)sr);
  Py_END_ALLOW_THREADS

  /* Does Disinfect() even trip callbacks? */
  if (PyErr_Occurred()) goto failed;
  if (self->exc_type) {
    PyErr_Restore(self->exc_type, self->exc_val, self->exc_tb);
    goto failed;
  }

  if (hr != SOPHOS_S_OK)
    return pysRaiseError(hr);

  Py_INCREF(Py_None);
  return Py_None;

failed:
  self->exc_type = self->exc_val = self->exc_tb = NULL;
  return NULL;
}

#define PYS_FUNCHANDLE(pfx)                                                   \
static PyObject *                                                             \
pys_I_##pfx##Handle(PySSavi3 *self, PyObject *args, PyObject *kwds) {         \
  static char *kwlist[] = { "handle", "name", "results", NULL };              \
  char *name = NULL;                                                          \
  PyObject *handle = NULL;                                                    \
  int want_sr = 1;                                                            \
  CIEnumSweepResults *pResults = NULL;                                        \
  HRESULT hr;                                                                 \
  int fd;                                                                     \
                                                                              \
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|zO&:" #pfx "Handle",        \
                                   kwlist, &handle, &name,                    \
                                   results_converter, &want_sr))              \
    return NULL;                                                              \
                                                                              \
  fd = PyObject_AsFileDescriptor(handle);                                     \
  if (fd == -1)                                                               \
    return NULL;                                                              \
                                                                              \
  Py_BEGIN_ALLOW_THREADS                                                      \
  hr = self->pSAVI->pVtbl->pfx##Handle(                                       \
          self->pSAVI,                                                        \
	        (LPCOLESTR) name,                                                   \
	        (SOPHOS_FD) fd,                                                     \
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                              \
          want_sr ? (void **)&pResults : NULL);                               \
  Py_END_ALLOW_THREADS                                                        \
                                                                              \
  return _pysavi_postSweep(self, hr, pResults);                               \
}

PYS_FUNCHANDLE(Sweep)
static char pys_I_SweepHandle__doc__[] = 
"SweepBuffer(buffer, name=None, results=SweepResults) -> viruses\n"
"\n"
"Scan the given file handle for viruses.  A file handle is either a whole\n"
"number (ala the file descriptor argument to os.fdopen) or an object\n"
"implementing a fileno() method (as ordinary file() objects).\n"
"\n"
"Return value is governed by the results parameter, as for SweepFile.\n";

PYS_FUNCHANDLE(Disinfect)
static char pys_I_DisinfectHandle__doc__[] = 
"DisinfectBuffer(buffer, name=None, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the file handle.  A file handle is either a whole\n"
"number (ala the file descriptor argument to os.fdopen) or an object\n"
"implementing a fileno() method (as ordinary file() objects).\n"
"\n"
"Return value is governed by the results parameter, as for DisinfectFile.\n";

#define PYS_FUNCBUFFER(pfx)                                                   \
static PyObject *                                                             \
pys_I_##pfx##Buffer(PySSavi3 *self, PyObject *args, PyObject *kwds) {         \
  static char *kwlist[] = { "buffer", "name", "results", NULL };              \
  char *buf; int buflen;                                                      \
  char *bufname = NULL;                                                       \
  int want_sr = 1;                                                            \
  CIEnumSweepResults *pResults = NULL;                                        \
  HRESULT hr;                                                                 \
                                                                              \
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#|zO&:" #pfx "Buffer",       \
                                   kwlist, &buf, &buflen, &bufname,           \
                                   results_converter, &want_sr))              \
    return NULL;                                                              \
                                                                              \
  Py_BEGIN_ALLOW_THREADS                                                      \
  hr = self->pSAVI->pVtbl->pfx##Buffer(                                       \
          self->pSAVI,                                                        \
	        (LPCOLESTR) bufname,                                                \
	        (U32) buflen,                                                       \
	        (U08 *)buf,                                                         \
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                              \
          want_sr ? (void **)&pResults : NULL);                               \
  Py_END_ALLOW_THREADS                                                        \
                                                                              \
  return _pysavi_postSweep(self, hr, pResults);                               \
}

PYS_FUNCBUFFER(Sweep)
static char pys_I_SweepBuffer__doc__[] = 
"SweepBuffer(buffer, name=None, results=SweepResults) -> viruses\n"
"\n"
"Scan the given buffer for viruses.  The name parameter, if supplied, is\n"
"used for identification purposes only (e.g., OnFileFound callbacks).\n"
"\n"
"Return value is governed by the results parameter, as for SweepFile.\n";

PYS_FUNCBUFFER(Disinfect)
static char pys_I_DisinfectBuffer__doc__[] = 
"DisinfectBuffer(buffer, name=None, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect buffer, and then scan it as SweepBuffer.  The name\n"
"parameter, if supplied, is used for identification purposes only (e.g.,\n"
"OnFileFound callbacks.\n"
"\n"
"Return value is governed by the results parameter, as for DisinfectFile.\n";

#define PYS_FUNCSTREAM(pfx)                                                   \
static PyObject *                                                             \
pys_I_##pfx##Stream(PySSavi3 *self, PyObject *args, PyObject *kwds) {         \
  static char *kwlist[] = { "stream", "name", "results", NULL };              \
  PyObject *stream = NULL;                                                    \
  char *name;                                                                 \
  void *pStream = NULL;                                                       \
  int want_sr = 1;                                                            \
  CIEnumSweepResults *pResults = NULL;                                        \
  HRESULT hr;                                                                 \
                                                                              \
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|zO&:" #pfx "Stream",        \
                                   kwlist, &stream, &name,                    \
                                   results_converter, &want_sr))              \
    return NULL;                                                              \
                                                                              \
  pStream = CIFileish_ctor(stream);                                           \
  if (!pStream) {                                                             \
    PyErr_SetString(PyExc_RuntimeError,                                       \
                    #pfx "Stream: internal SAVISTREAM creation failed");      \
    return NULL;                                                              \
  }                                                                           \
                                                                              \
  Py_BEGIN_ALLOW_THREADS                                                      \
  hr = self->pSAVI->pVtbl->pfx##Stream(                                       \
          self->pSAVI,                                                        \
	        (LPCOLESTR) name,                                                   \
          (REFIID)&SOPHOS_IID_SAVISTREAM,                                     \
	        (void *) pStream,                                                   \
          (REFIID)&SOPHOS_IID_ENUM_SWEEPRESULTS,                              \
          want_sr ? (void **)&pResults : NULL);                               \
  Py_END_ALLOW_THREADS                                                        \
  pysIThingSafeRelease(((CISweepUnknown2 *)pStream));                         \
                                                                              \
  return _pysavi_postSweep(self, hr, pResults);                               \
}

PYS_FUNCSTREAM(Sweep)
static char pys_I_SweepStream__doc__[] = 
"SweepStream(stream, name=None, results=SweepResults) -> viruses\n"
"\n"
"Scan the given stream for viruses.  The stream object must implement\n"
"file-like methods read(), seek() and tell().  StringIO objects, for example,\n"
"are streams in this sense.  The optional name parameter is used for\n"
"identification purposes (e.g., OnFileFound callbacks).\n"
"\n"
"Return value is governed by the results parameter, as for SweepFile.\n";

PYS_FUNCSTREAM(Disinfect)
static char pys_I_DisinfectStream__doc__[] =
"DisinfectStream(stream, name=None, results=SweepResults) -> viruses\n"
"\n"
"Attempt to disinfect the passed stream.  The stream object must implement\n"
"file-like methods read(), write(), seek() and tell().  Writeable StringIO\n"
"objects, for example, are streams in this sense.  The optional name\n"
"parameter is used for identification purposes (e.g., OnFileFound callbacks).\n"
"\n"
"Return value and results parameter are as for DisinfectFile.\n";


static char pys_I_GetVirusEngineVersion__doc__[] =
"GetVirusEngineVersion(details=None) -> misc_tuple\n"
"\n"
"Returns a tuple miscellaneous information about the virus engine.  Tuple\n"
"elements are: integral engine version, string version, virus data date (not\n"
"implemented), number of detectable viruses, None, and data file details.\n"
"\n"
"The data file details element depends on the details parameter.  If None,\n"
"None is returned.  If the IDEDetails or VersionChecksum type is supplied, a\n"
"corresponding iterator is returned.\n";

/* Warning:  repellent C function wrapped in repellent python garb... */
static PyObject *
pys_I_GetVirusEngineVersion(PySSavi3 *self, PyObject *args, PyObject *kwds) {

  static char *kwlist[] = { "details", NULL };
  PyObject *details = NULL;
  /* -- */
  OLECHAR ver_str[256];
  U32 version, num_detectable;
  REFIID DetailsIID;
  CIEnumGeneric *pEnum = NULL;
  PyObject *det_param = NULL;
  HRESULT hr;

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O:GetVirusEngineVersion",
                                   kwlist, &details))
    return NULL;

  if (NULL == details || details == Py_None) {
    DetailsIID = &SOPHOS_IID_ENUM_IDEDETAILS;
    details = NULL; /* meaning: skip pEnum */
  } else if (details == (PyObject *)&PySIDEDetails_Type) {
    DetailsIID = &SOPHOS_IID_ENUM_IDEDETAILS;
  } else if (details == (PyObject *)&PySVersionChecksum_Type) {
    DetailsIID = &SOPHOS_IID_ENUM_CHECKSUM;
  } else {
    PyErr_SetString(PyExc_TypeError,
                    "Invalid or unrecognized details parameter");
    return NULL;
  }
  /* FIXME - error check det_param */

  Py_BEGIN_ALLOW_THREADS
  hr = self->pSAVI->pVtbl->GetVirusEngineVersion(
         self->pSAVI,
         &version,
         ver_str,
         sizeof(ver_str) -1,
         (SYSTEMTIME*) NULL,
         &num_detectable,
         (U32*) NULL,
         DetailsIID,
         details ? (void **)&pEnum : NULL);
  Py_END_ALLOW_THREADS

  if (SOPHOS_FAILED(hr)) {
    pysIThingSafeRelease(pEnum);
    return pysRaiseError(hr);
  }

  if (! details) {
    det_param = Py_None;
    Py_INCREF(det_param);
  } else {
    det_param = PySEnumGeneric_New(DetailsIID, pEnum, 1);
  }
  return Py_BuildValue("lsslsN", (long) version, ver_str, NULL,
                                 (long) num_detectable, NULL,
                                 det_param);
}

static char pys_I_LoadVirusData__doc__[] =
"LoadVirusData()\n"
"\n"
"Release any existing virus data and reload virus definitions.  This method,\n"
"called implicitly by the first call to a Sweep...() function or to\n"
"GetVirusEngineVersion(), may invoke an OnChange callback (if any).\n"
"\n"
"This method will raise an exception if the loaded virus data is too old.  To\n"
"(foolishly) ignore this error, but not other errors, the caller may try:\n"
"\n"
"    sav = pysavi.Savi3()\n"
"    try:\n"
"        sav.LoadVirusData()\n"
"    except pysavi.error, e:\n"
"        if e[0] <> pysavi.SOPHOS_SAVI_ERROR_OLD_VIRUS_DATA: raise\n"
"\n";

static PyObject *
pys_I_LoadVirusData(PySSavi3 *self) {  /* METH_NOARGS */
  HRESULT hr;

  Py_BEGIN_ALLOW_THREADS
  hr = self->pSAVI->pVtbl->LoadVirusData(self->pSAVI);
  Py_END_ALLOW_THREADS

  if (SOPHOS_FAILED(hr) || hr == SOPHOS_SAVI_ERROR_OLD_VIRUS_DATA)
    return pysRaiseError(hr);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyMethodDef pys_I_methods[] = { 
  {"SweepFile",     (PyCFunction)pys_I_SweepFile,     METH_KEYWORDS, 
    pys_I_SweepFile__doc__},
  {"DisinfectFile", (PyCFunction)pys_I_DisinfectFile, METH_KEYWORDS, 
    pys_I_DisinfectFile__doc__},
  
  {"SweepHandle",     (PyCFunction)pys_I_SweepHandle,     METH_KEYWORDS, 
    pys_I_SweepHandle__doc__},
  {"DisinfectHandle", (PyCFunction)pys_I_DisinfectHandle, METH_KEYWORDS, 
    pys_I_DisinfectHandle__doc__},

  {"SweepStream",     (PyCFunction)pys_I_SweepStream,     METH_KEYWORDS, 
    pys_I_SweepStream__doc__},
  {"DisinfectStream", (PyCFunction)pys_I_DisinfectStream, METH_KEYWORDS, 
    pys_I_DisinfectStream__doc__},

  {"SweepBuffer",     (PyCFunction)pys_I_SweepBuffer,     METH_KEYWORDS, 
    pys_I_SweepBuffer__doc__},
  {"DisinfectBuffer", (PyCFunction)pys_I_DisinfectBuffer, METH_KEYWORDS, 
    pys_I_DisinfectBuffer__doc__},

  {"SweepLogicalSector",     (PyCFunction)pys_I_SweepLogicalSector,
     METH_KEYWORDS, pys_I_SweepLogicalSector__doc__},
  {"DisinfectLogicalSector", (PyCFunction)pys_I_DisinfectLogicalSector,
     METH_KEYWORDS, pys_I_DisinfectLogicalSector__doc__},

  {"SweepPhysicalSector",     (PyCFunction)pys_I_SweepPhysicalSector,
     METH_KEYWORDS, pys_I_SweepPhysicalSector__doc__},
  {"DisinfectPhysicalSector", (PyCFunction)pys_I_DisinfectPhysicalSector,
     METH_KEYWORDS, pys_I_DisinfectPhysicalSector__doc__},

  {"SweepMemory", (PyCFunction)pys_I_SweepMemory, METH_KEYWORDS, 
    pys_I_SweepMemory__doc__},
  {"Disinfect", (PyCFunction)pys_I_Disinfect, METH_VARARGS, 
    pys_I_Disinfect__doc__},
  {"LoadVirusData", (PyCFunction)pys_I_LoadVirusData, METH_NOARGS, 
    pys_I_LoadVirusData__doc__},
  {"SetConfigDefaults", (PyCFunction)pys_I_SetConfigDefaults, METH_NOARGS, 
    pys_I_SetConfigDefaults__doc__},
  {"SetConfigValue", (PyCFunction)pys_I_SetConfigValue, METH_VARARGS, 
    pys_I_SetConfigValue__doc__},
  {"GetVirusEngineVersion", (PyCFunction)pys_I_GetVirusEngineVersion,
     METH_KEYWORDS, pys_I_GetVirusEngineVersion__doc__},
  {"GetConfigValue", (PyCFunction)pys_I_GetConfigValue, METH_VARARGS, 
    pys_I_GetConfigValue__doc__},
  {"GetConfigEnumerator", (PyCFunction)pys_I_GetConfigEnumerator, METH_NOARGS,
     pys_I_GetConfigEnumerator__doc__},
  {"RegisterNotification", (PyCFunction)pys_I_RegisterNotification,
     METH_VARARGS, pys_I_RegisterNotification__doc__},
  { NULL,   NULL }
};

#undef METHOD_ENTRY

statichere PyTypeObject PySSavi3_Type = {
        /* The ob_type field must be initialized in the module init function
         * to be portable to Windows without using C++. */
        PyObject_HEAD_INIT(NULL)
        0,                      /*ob_size*/
        "_savi.Savi3",         /*tp_name*/
        sizeof(PySSavi3),         /*tp_basicsize*/
        0,                      /*tp_itemsize*/
        /* methods */
        (destructor)pys_I_dealloc, /*tp_dealloc*/
        0,                      /*tp_print*/
        0,                      /*tp_getattr*/
        0,                      /*tp_setattr*/
        0,                      /*tp_compare*/
        0,                      /*tp_repr*/
        0,                      /*tp_as_number*/
        0,                      /*tp_as_sequence*/
        0,                      /*tp_as_mapping*/
        0,                      /*tp_hash*/
        0,                      /*tp_call*/
        0,                      /*tp_str*/
        PyObject_GenericGetAttr, /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,/*tp_flags*/
        PySSavi3__doc__,          /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        pys_I_methods,          /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        0,                      /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        (initproc) pys_I_init,  /*tp_init*/
        0,                      /*tp_alloc*/
        pys_I_new,              /*tp_new*/
        _PyObject_Del,          /*tp_free*/
        0,                      /*tp_is_gc*/
};

/* ---------------- Module -------------------- */

static PyMethodDef module_functions[] = {
  {"strerror", (PyCFunction)pysavi_strerror, METH_VARARGS,
    pysavi_strerror__doc__},
  {NULL,  NULL}
};

static char module_doc[] =
"Lower-level interface to the Sophos ISavi (SAVI2) functions\n";

static void
pys_export_type(PyObject *module, PyTypeObject *t) {
  char *n;

  n = strrchr(t->tp_name, '.');
  if (!n) n = t->tp_name;
  else    n = n + 1;

  Py_INCREF((PyObject *)t);
  PyModule_AddObject(module, n, (PyObject *)t); 
}

DL_EXPORT(void)
init_savi(void) {
  PyObject *m, *d, *v;

  Py_Initialize();
  PyEval_InitThreads();


#define READY_TYPE(T)                 \
  (T).ob_type = &PyType_Type;         \
  if (PyType_Ready(&(T)) < 0) return;

  READY_TYPE(PySSavi3_Type);
  READY_TYPE(PySSweepResults_Type);
  READY_TYPE(PySSweepError_Type);
  READY_TYPE(PySIDEDetails_Type);
  READY_TYPE(PySEngineConfig_Type);
  READY_TYPE(PySVersionChecksum_Type);
  READY_TYPE(PySEnumGeneric_Type);

  m = Py_InitModule3("_savi", module_functions, module_doc);
  pysInitModuleConstants(m);

  d = PyModule_GetDict(m);

  PyDict_SetItemString(d, "__version__", 
                       v = PyString_FromString(VERSION));
  Py_XDECREF(v);

  PyDict_SetItemString(d, "__revision__", 
                       v = PyString_FromString(REVISION));
  Py_XDECREF(v);

  pysErrorObject = PyErr_NewException("pysavi.error", PyExc_EnvironmentError, NULL);
  PyDict_SetItemString(d, "error", pysErrorObject);

  pys_export_type(m, &PySSavi3_Type);
  Py_INCREF((PyObject *)&PySSavi3_Type);
  PyModule_AddObject(m, (char *)"Savi", (PyObject *)&PySSavi3_Type);
  pys_export_type(m, &PySSweepResults_Type);
  pys_export_type(m, &PySSweepError_Type);
  pys_export_type(m, &PySIDEDetails_Type);
  pys_export_type(m, &PySEngineConfig_Type);
  pys_export_type(m, &PySVersionChecksum_Type);
  pys_export_type(m, &PySEnumGeneric_Type);
}

/* vim:et:ts=2:noai
 */
