/* $Id: pysaviCOM.h,v 1.1.1.1 2004/11/02 03:57:58 mjp Exp $
 *
 * pysaviCOM.h
 *
 * Helper macros to define QueryInterface, AddRef and Release for pysavi
 * COM-style interfaces.  Note that Release calls PyMem_Free() and performs
 * no other actions on the interface object.
 */

#ifndef __PYSAVICOM_H__
#define __PYSAVICOM_H__
#include "Python.h"
#include "csavi3c.h"

#define PYSCOM_QUERYINTERFACE(PFX, TYPE, IID)                               \
static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC                                 \
PFX##_QueryInterface(void *object, REFIID IIDObject, void **ppObject) {     \
  TYPE *INTobject = (TYPE *)object;                                         \
  HRESULT             retCode;                                              \
                                                                            \
  if (!object || !IIDObject || !ppObject ||                                 \
      !IsEqualIID(&INTobject->typeCode, &(IID)))                            \
    return SOPHOS_E_INVALIDARG;                                             \
                                                                            \
  if (!IsEqualIID(IIDObject, &SOPHOS_IID_UNKNOWN2) &&                       \
      !IsEqualIID(IIDObject, &INTobject->typeCode))                         \
    return SOPHOS_E_NOINTERFACE;                                            \
                                                                            \
  retCode = ((SOPHOS_FAILED(INTobject->pVtbl->AddRef(object)))              \
            ? SOPHOS_E_UNEXPECTED : SOPHOS_S_OK);                           \
                                                                            \
  /*                                                                        \
  printf("QueryInterface(%p, %s)\n", object, #IID);                         \
  */                                                                        \
                                                                            \
  *ppObject = ((retCode == SOPHOS_S_OK) ? object : NULL);                   \
  return retCode;                                                           \
}

#define PYSCOM_ADDREF(PFX, TYPE, IID)                                       \
static SOPHOS_ULONG SOPHOS_STDCALL SOPHOS_PUBLIC                            \
PFX##_AddRef(void *object) {                                                \
  TYPE *INTobject = (TYPE *)object;                                         \
                                                                            \
  if (!object ||                                                            \
      !IsEqualIID(&INTobject->typeCode, &(IID)))                            \
    return (SOPHOS_ULONG)SOPHOS_E_INVALIDARG;                               \
                                                                            \
  if (INTobject->refCount == 0x7FFFFFFF)                                    \
    return (SOPHOS_ULONG)SOPHOS_E_UNEXPECTED;                               \
                                                                            \
  /*                                                                        \
  printf("AddRef(%p)\n", object);                                           \
  */                                                                        \
                                                                            \
  return (++(INTobject->refCount));                                         \
}

/* Release() - DTOR is called with one argument, the object */
#define PYSCOM_RELEASE(PFX, TYPE, IID, DTOR)                                \
static SOPHOS_ULONG SOPHOS_STDCALL SOPHOS_PUBLIC                            \
PFX##_Release(void *object) {                                               \
  TYPE *INTobject = (TYPE *)object;                                         \
  SOPHOS_ULONG      retCode;                                                \
                                                                            \
  if (!object ||                                                            \
      !IsEqualIID(&INTobject->typeCode, &(IID)))                            \
    return (SOPHOS_ULONG)SOPHOS_E_INVALIDARG;                               \
                                                                            \
  if (INTobject->refCount > 0)                                              \
    INTobject->refCount--;                                                  \
                                                                            \
  retCode = INTobject->refCount;                                            \
                                                                            \
  /*                                                                        \
  printf(#IID "->Release(%p) (refcnt %lu)\n", object, retCode);             \
  */                                                                        \
                                                                            \
  if (retCode == 0)                                                         \
    DTOR((void *)object);                                                   \
                                                                            \
  return retCode;                                                           \
}

#endif /* __PYSAVICOM_H__ */

/* vim:et:ts=2:noai
 */
