/* $Id */
#include "Python.h"
#include "csavi3c.h"
#include "pysavi.h"
#include "pysaviCOM.h"
#include "pysaviTypes.h"
#include "pysaviUtil.h"

typedef struct _CICbckGeneric {
  CISweepUnknownVtbl  *pVtbl;        /* QueryInterface, AddRef, Release */
  IID                 typeCode;
  U32                 refCount;
} CICbckGeneric;

/* -- predeclarations: misc -- */

int pysCbckRegister(PySSavi3 *, REFIID);

static HRESULT
  pysCbckDispatch(PyObject *token, char *name, char *fmt, ...);

/* -- SaviNotify implementation -- */

/* CISweepNotify2 objects have dynamically allocated vtbls (our other COM
 * interfaces share static vtbls), so our DTOR needs to free two things.
 *
 * Note - when we're called, type checking and reference counting have already
 * been done.
 */
static void
_SaviNotify_dtor(void *o) {
  CISweepNotify2 *INTobject = (CISweepNotify2 *)o;
  PyMem_Free(INTobject->pVtbl);
  PyMem_Free(o);
}

PYSCOM_QUERYINTERFACE(SaviNotify, CISweepNotify2, SOPHOS_IID_SWEEPNOTIFY2)
PYSCOM_ADDREF(SaviNotify, CISweepNotify2, SOPHOS_IID_SWEEPNOTIFY2)
/*
PYSCOM_RELEASE(SaviNotify, CISweepNotify2, SOPHOS_IID_SWEEPNOTIFY2, PyMem_Free)
*/
PYSCOM_RELEASE(SaviNotify, CISweepNotify2, SOPHOS_IID_SWEEPNOTIFY2, _SaviNotify_dtor)

static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviNotify_OnFileFound(void *object, void *token, LPCOLESTR pName) {
  if (!object || !token) return SOPHOS_E_INVALIDARG;

  return pysCbckDispatch((PyObject *)token, "OnFileFound",
                         "(s)", (char *)pName);
}

static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviNotify_OnVirusFound(void *object, void *token, REFIID ResultsIID,
                        void *pResults) {
  CISweepNotify2 *INTobject  = (CISweepNotify2 *)object;
  CISweepResults *INTresults = (CISweepResults *)pResults;

  if (!object || !pResults || 
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_SWEEPNOTIFY2) ||
      !IsEqualIID(ResultsIID, &SOPHOS_IID_SWEEPRESULTS))
    return SOPHOS_E_INVALIDARG;

  return pysCbckDispatch((PyObject *)token, "OnVirusFound",
                         "(N)", PySSweepResults_New(INTresults));
}

static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviNotify_OnErrorFound(void *object, void *token, REFIID ErrorIID,
                        void *pError) {
  CISweepNotify2 *INTobject  = (CISweepNotify2 *)object;
  CISweepError *INTerror = (CISweepError *)pError;

  if (!object || !pError ||
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_SWEEPNOTIFY2) ||
      !IsEqualIID(ErrorIID, &SOPHOS_IID_SWEEPERROR))
    return SOPHOS_E_INVALIDARG;

  return pysCbckDispatch((PyObject *)token, "OnErrorFound",
                         "(N)", PySSweepError_New(INTerror));
}

static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviNotify_OnClassification(void *object, void *token, U32 Classifn) {
  CISweepNotify2 *INTobject  = (CISweepNotify2 *)object;

  if (!object || !token ||
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_SWEEPNOTIFY2))
    return SOPHOS_E_INVALIDARG;

  return pysCbckDispatch((PyObject *)token, "OnClassification",
                         "(i)", Classifn);
}
                
static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviNotify_OkToContinue(void *object, void *token, U16 Activity, U32 Extent,
                        LPCOLESTR pTarget) {
  CISweepNotify2 *INTobject  = (CISweepNotify2 *)object;

  if (!object || !token ||
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_SWEEPNOTIFY2))
    return SOPHOS_E_INVALIDARG;

  return pysCbckDispatch((PyObject *)token, "OkToContinue",
                         "(iis)", Activity, Extent, pTarget);
}

static const CISweepNotify2Vtbl CISaviNotifyVtbl_Instance = {
  SaviNotify_QueryInterface,
  SaviNotify_AddRef,
  SaviNotify_Release,
  SaviNotify_OnFileFound,
  SaviNotify_OnVirusFound,
  SaviNotify_OnErrorFound,
  SaviNotify_OkToContinue,
  SaviNotify_OnClassification
};

/* -- SaviChange implementation -- */

PYSCOM_QUERYINTERFACE(SaviChange, CIChangeNotify, SOPHOS_IID_CHANGENOTIFY)
PYSCOM_ADDREF(SaviChange, CIChangeNotify, SOPHOS_IID_CHANGENOTIFY)
PYSCOM_RELEASE(SaviChange, CIChangeNotify, SOPHOS_IID_CHANGENOTIFY, PyMem_Free)

static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
SaviChange_OnChange(void *object, void *token) {
  CIChangeNotify *INTobject = (CIChangeNotify *)object;

  if (!object || !token ||
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_CHANGENOTIFY))
    return SOPHOS_E_INVALIDARG;

  (void) pysCbckDispatch((PyObject *)token, "OnChange", "()");
  return (HRESULT)0; /* ignored */
}

static const CIChangeNotifyVtbl CIChangeNotifyVtbl_Instance = {
  SaviChange_QueryInterface,
  SaviChange_AddRef,
  SaviChange_Release,
  SaviChange_OnChange
};

/* -- SaviSevereError implementation -- */

PYSCOM_QUERYINTERFACE(SaviSevereError, CISeverityNotify, SOPHOS_IID_SEVERITYNOTIFY)
PYSCOM_ADDREF(SaviSevereError, CISeverityNotify, SOPHOS_IID_SEVERITYNOTIFY)
PYSCOM_RELEASE(SaviSevereError, CISeverityNotify, SOPHOS_IID_SEVERITYNOTIFY, PyMem_Free)

/* def OnSevereError(self, errCode, severity)
 *
 * rtrn value ignored, exceptions masked.
 */
static HRESULT SOPHOS_STDCALL SOPHOS_PUBLIC
  SaviSevereError_OnSevereError(void *object,
                                void *token, HRESULT Err, U32 Severity) {
  CISeverityNotify *INTobject = (CISeverityNotify *)object;

  if (!object || !token ||
      !IsEqualIID(&INTobject->typeCode, &SOPHOS_IID_SEVERITYNOTIFY))
    return SOPHOS_E_INVALIDARG;

  (void) pysCbckDispatch((PyObject *)token, "OnSevereError",
                         "(ii)", Err, Severity);
  return (HRESULT)0; /* ignored */
}

static const CISeverityNotifyVtbl CISaviSevereErrorVtbl_Instance = {
  SaviSevereError_QueryInterface,
  SaviSevereError_AddRef,
  SaviSevereError_Release,
  SaviSevereError_OnSevereError,
};

/* -- Misc Functions -- */

/* pysCbckRegister
 *
 * Helper function to create COM interfaces for callback registration,
 * allocating vtbls if necessary.  Assumes that the "cbkc" member of the
 * PySSavi3 object is valid; this function does *not* alter any python
 * refcounts.
 */
int
pysCbckRegister(PySSavi3 *pys, REFIID riid) {
  HRESULT hr;
  size_t malloc_sz = 0;
  CICbckGeneric *INTobject = NULL;
  CISweepUnknownVtbl *pVtbl = NULL;

  if (IsEqualIID(riid, &SOPHOS_IID_SWEEPNOTIFY2)) {
    /* SaviNotify, unlike the others, has a dynamic vtbl per instance so we
     * only slot methods implemented at the time of registration.
     *
     * 1) Allocate vtbl
     * 2) Slot methods (pyobj.OnFileFound -> vtbl->OnFileFound)
     * 3) Allocate COM interface thingy (as normal)
     */
    CISweepNotify2Vtbl *v = (CISweepNotify2Vtbl *)PyMem_Malloc(sizeof(CISweepNotify2Vtbl));
    memset(v, '\0', sizeof(CISweepNotify2Vtbl));

    v->QueryInterface = SaviNotify_QueryInterface;
    v->AddRef = SaviNotify_AddRef;
    v->Release = SaviNotify_Release;
    if (pysPyObjHasMethod(pys->cbck, "OnFileFound"))
      v->OnFileFound = SaviNotify_OnFileFound;
    if (pysPyObjHasMethod(pys->cbck, "OnVirusFound"))
      v->OnVirusFound = SaviNotify_OnVirusFound;
    if (pysPyObjHasMethod(pys->cbck, "OnErrorFound"))
      v->OnErrorFound = SaviNotify_OnErrorFound;
    if (pysPyObjHasMethod(pys->cbck, "OkToContinue"))
      v->OkToContinue = SaviNotify_OkToContinue;
    if (pysPyObjHasMethod(pys->cbck, "OnClassification"))
      v->OnClassification = SaviNotify_OnClassification;

    malloc_sz = sizeof(CISweepNotify2);
    pVtbl = (CISweepUnknownVtbl *)v;
  } else if (IsEqualIID(riid, &SOPHOS_IID_CHANGENOTIFY)) {
    malloc_sz = sizeof(CIChangeNotify);
    pVtbl = (CISweepUnknownVtbl *)&CIChangeNotifyVtbl_Instance;
  } else if (IsEqualIID(riid, &SOPHOS_IID_SEVERITYNOTIFY)) {
    malloc_sz = sizeof(CISeverityNotify);
    pVtbl = (CISweepUnknownVtbl *)&CISaviSevereErrorVtbl_Instance;
  } else {
    pysRaiseError(SOPHOS_E_NOINTERFACE);
    return 0;
  }

  /* Allocate COM interface */
  INTobject = (CICbckGeneric *)PyMem_Malloc(malloc_sz);
  if (!INTobject) return 0;
  INTobject->pVtbl    = pVtbl;
  INTobject->typeCode = *riid;
  INTobject->refCount = 1;

  hr = pys->pSAVI->pVtbl->RegisterNotification(pys->pSAVI,
                                               riid,
                                               (void *)INTobject,
                                               (void *)pys);
  pysIThingSafeRelease(INTobject);
  if (!SOPHOS_SUCCEEDED(hr)) {
    pysRaiseError(hr);
    return 0;
  }
  return 1;
}

/* pysCbckDispatch(PyObject *token, char *name, char *fmt, ...)
 *
 * Acquire the GIL, dispatch the callback as a method call of the registered
 * object, and interpret its return value.  
 *
 * token - the PySSavi3 object which registered the callback
 *         token->cbck is the object offering the methods
 *         token->interp is the interpreter state
 *         
 *  name - method name to call
 *
 *   fmt - Py_BuildValue format and args
 *   ...
 */

static HRESULT
pysCbckDispatch(PyObject *token, char *name, char *fmt, ...) {
  va_list va;
  HRESULT hr;
  PyThreadState *tstate;
  PyObject *args, *ret, *func = NULL;
  PySSavi3 *self = (PySSavi3 *)token;

  tstate = PyThreadState_New(self->interp); /*THREADS!*/
  PyEval_AcquireThread(tstate);

  func = PyObject_GetAttrString(self->cbck, name);
  if (NULL == func) { 
    hr = SOPHOS_SAVI2_ERROR_CALLBACK;
    goto out;
  }

  if (fmt && *fmt) {
    va_start(va, fmt);
    args = Py_VaBuildValue(fmt, va);
    va_end(va);
  }
  else
    args = PyTuple_New(0);

  if (NULL == args) goto unexpected;

  if (!PyTuple_Check(args)) {
    PyObject *a;

    a = PyTuple_New(1);
    if (NULL == a) goto unexpected;
    if (PyTuple_SetItem(a, 0, args) < 0)
      goto unexpected;
    args = a;
  }

  ret = PyObject_CallObject(func, args);
  if (NULL == ret) goto unexpected;

  if (ret == Py_None) {
    hr = SOPHOS_SAVI_CBCK_DEFAULT;
  } else if (PyInt_Check(ret))
    hr = (HRESULT) PyInt_AsLong(ret);
  else {
    hr = SOPHOS_E_UNEXPECTED;
  }
  Py_DECREF(ret);

out:
  Py_XDECREF(func);
  PyEval_ReleaseThread(tstate);
  DBG("pysCbckDispatch('%s') returning %ld\n", name, hr);
  return hr;

unexpected:
  /* callback threw an exception, or our internal py mechanics did */
  DBG("\tpysCbckDispatch('%s') inside 'unexpected branch'\n", name);
  if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_RuntimeError, "Unexpected exception in callback");
  } else {
    PyErr_Fetch(&self->exc_type, &self->exc_val, &self->exc_tb);
  }
  hr = SOPHOS_SAVI_CBCK_STOP;
  goto out;
}


/* vim:et:ts=2:noai
 */
