/* $Id: pysaviUtil.c,v 1.2 2005/02/08 06:06:14 mjp Exp $ */

#include <Python.h>
#include "csavi3c.h"

/* pys_strerror(HRESULT)
 *
 */
char *
pysStrerror(HRESULT hr) {
  int i;
  char *errmsg = "Unknown error";
  static struct { 
      HRESULT hr;
      char *errmsg;
  } sophos_errors[] = {
    { SOPHOS_S_OK, "Success" },
    { SOPHOS_S_FALSE, "Partial Success" },
    { SOPHOS_E_NOINTERFACE, "Unknown or unsupported interface" },
    { SOPHOS_CLASS_E_NOAGGREGATION, "Invalid ClassFactory parameter" },
    { SOPHOS_E_UNEXPECTED, "Unexpected error occurred" },
    { SOPHOS_E_OUTOFMEMORY, "Out of memory" },
    { SOPHOS_E_INVALIDARG, "Invalid argument" },
    { SOPHOS_E_NOTIMPL, "Function not implemented" },
    { SOPHOS_E_OUT_OF_DISK, "File creation or write problem" },
    { SOPHOS_RPC_E_WRONG_THREAD, "Callback installed in different thread" },
    { SOPHOS_SAVI_ERROR_INITIALISING, "SAVI interface could not be initialised" },
    { SOPHOS_SAVI_ERROR_TERMINATING, "SAVI interface could not be terminated" },
    { SOPHOS_SAVI_ERROR_SWEEPFAILURE, "Error occurred during virus scan" },
    { SOPHOS_SAVI_ERROR_VIRUSPRESENT, "Virus found during virus scan" },
    { SOPHOS_SAVI_ERROR_NOT_INITIALISED, "Function called prior to interface initialisation" },
    { SOPHOS_SAVI_ERROR_IC_INCOMPATIBLE_VERSION, "InterCheck version incompatible with SAVI" },
    { SOPHOS_SAVI_ERROR_IC_ACCESS_DENIED, "Insufficient rights to disable InterCheck" },
    { SOPHOS_SAVI_ERROR_IC_SCAN_PREVENTED, "Unable to disable InterCheck" },
    { SOPHOS_SAVI_ERROR_DISINFECTION_FAILED, "Disinfection failed" },
    { SOPHOS_SAVI_ERROR_DISINFECTION_UNAVAILABLE, "Disinfection unavailable" },
    { SOPHOS_SAVI_ERROR_UPGRADE_FAILED, "Upgrade (setup) failed" },
    { SOPHOS_SAVI_ERROR_SAV_NOT_INSTALLED, "Sophos Anti-Virus not found" },
    { SOPHOS_SAVI_ERROR_INVALID_CONFIG_NAME, "Invalid configuration setting name" },
    { SOPHOS_SAVI_ERROR_INVALID_CONFIG_TYPE, "Invalid configuration setting type" },
    { SOPHOS_SAVI_ERROR_INIT_CONFIGURATION, "Internal error encountered accessing configuration" },
    { SOPHOS_SAVI_ERROR_NOT_SUPPORTED, "Unrecognized file type or unsupported function" },
    { SOPHOS_SAVI_ERROR_COULD_NOT_OPEN, "Could not access item" },
    { SOPHOS_SAVI_ERROR_FILE_COMPRESSED, "No virus found at outer level of compressed file" },
    { SOPHOS_SAVI_ERROR_FILE_ENCRYPTED, "Cannot scan encrypted file" },
    { SOPHOS_SAVI_ERROR_INFORMATION_NOT_AVAILABLE, "Additional virus location unavailable" },
    { SOPHOS_SAVI_ERROR_ALREADY_INIT, "SAVI interface already initialized" },
    { SOPHOS_SAVI_ERROR_STUB, "Attempted to use stub library" },
    { SOPHOS_SAVI_ERROR_BUFFER_TOO_SMALL, "Supplied buffer too small" },
    { SOPHOS_SAVI_ERROR_CORRUPT, "Cannot scan corrupt file" },
    { SOPHOS_SAVI_ERROR_REENTRANCY, "Prohibited SAVI re-entrance attempt from callback" },
    { SOPHOS_SAVI_ERROR_CALLBACK, "Error occurred using callback" },
    { SOPHOS_SAVI_ERROR_PARTIAL_INFORMATION, "GetVirusEngineVersion() succeeded partially" },
    { SOPHOS_SAVI_ERROR_OLD_VIRUS_DATA, "Internal virus data old, update needed" },
    { SOPHOS_SAVI_ERROR_INVALID_TMP, "No valid temp directory found" },
    { SOPHOS_SAVI_ERROR_MISSING_MAIN_VIRUS_DATA, "Main virus data missing" },
    { SOPHOS_SAVI_INFO_IC_ACTIVE, "Could not disable active InterCheck client" },
    { SOPHOS_SAVI_ERROR_VIRUS_DATA_INVALID_VER, "Invalid version of main virus data" },
    { SOPHOS_SAVI_ERROR_MUST_REINIT, "Must reinitialise, virus engine postdates SAVI" },
    { SOPHOS_SAVI_ERROR_CANNOT_SET_OPTION, "Cannot set configuration option" },
    { SOPHOS_SAVI_ERROR_PART_VOL, "Cannot scan part of multi-volume archive" },
    { SOPHOS_SAVI_INFO_OPT_GRP_INVAL_RTN, "Meaningless attempt to interrogate group option setting" },
    { SOPHOS_SAVI_CBCK_DEFAULT, "Default callback return value" },
    { SOPHOS_SAVI_ERROR_VDLD_ACTIVITY, "Cannot scan during virus data operation" },
    { SOPHOS_SAVI_ERROR_STREAM_READ_FAIL, "Stream read failed" },
    { SOPHOS_SAVI_ERROR_STREAM_WRITE_FAIL, "Stream write failed" },
    { SOPHOS_SAVI_ERROR_STREAM_SEEK_FAIL, "Stream seek failed" },
    { SOPHOS_SAVI_ERROR_STREAM_GETLENGTH_FAIL, "Stream length determination failed" },
    { SOPHOS_SAVI_ERROR_MISSING_VDL_PART, "Split virus data set error" },
    { SOPHOS_SAVI_WARNING_MISSING_VDL_PART, "Warning: Split virus data set error" },
    { SOPHOS_SAVI_ERROR_VDL_CHECKSUM, "Wrong checksum detected" },
    { SOPHOS_SAVI_WARNING_VDL_CHECKSUM, "Warning: Wrong checksum detected" },
    { SOPHOS_SAVI_ERROR_SCAN_ABORTED, "AutoStop aborted scan" },
  };

  for (i = 0; i < sizeof(sophos_errors)/sizeof(sophos_errors[0]); i++) {
    if (hr == sophos_errors[i].hr) {
      errmsg = sophos_errors[i].errmsg;
    }
  }

  return errmsg;
}

PyObject *
pysRaiseError(HRESULT hr) {
  extern PyObject *pysErrorObject;
  PyObject *v;

  v = Py_BuildValue("(is)", (long)hr, pysStrerror(hr));
  if (NULL != v) {
    PyErr_SetObject(pysErrorObject, v);
    Py_DECREF(v);
  }

  return NULL;
}

int
pysPyObjHasMethod(PyObject *o, const char *mname) {
  PyObject *m;
  int ret = 0;

  if (! PyObject_HasAttrString(o, (char *)mname)) return ret;

  m = PyObject_GetAttrString(o, (char *)mname);
  if (m == NULL) {
    /* Uh, but HasAttrString just succeeded */
    return 0;
  }

  ret = PyCallable_Check(m);
  Py_DECREF(m);

  return ret;
}

PyObject *
pysIThingGetString(void *thing, HRESULT (SOPHOS_STDCALL_PUBLIC_PTR f)(void *, U32, LPOLESTR, U32*)) {
  HRESULT hr;
  U32 len;
  PyObject *ret = NULL;

  hr = (*f)(thing, 0, NULL, &len);
  if (!SOPHOS_SUCCEEDED(hr)) return pysRaiseError(hr);

  /* len includes terminator, and we know python makes room for an extra
   * '\0' under the hood ...
   * Ugly.
   */
  ret = PyString_FromStringAndSize(NULL, len - 1);
  if (ret == NULL) return NULL;

  hr = (*f)(thing, len, (LPOLESTR)PyString_AsString(ret), NULL);
  if (!SOPHOS_SUCCEEDED(hr)) {
    Py_DECREF(ret);
    return pysRaiseError(hr);
  }

  return ret;
}

PyObject *
pysIThingGetInt(void *thing, HRESULT (SOPHOS_STDCALL_PUBLIC_PTR f)(void *, U32 *)) {
  HRESULT hr;
  U32 i;

  hr = (*f)(thing, &i);
  if (!SOPHOS_SUCCEEDED(hr)) return pysRaiseError(hr);
  return PyInt_FromLong(i);
}

